@@ -16,16 +16,20 @@
 package agentendpoint
 
 import (
+	"bytes"
 	"context"
+	"crypto/sha256"
+	"encoding/hex"
 	"errors"
 	"fmt"
 	"io"
 	"sync"
 	"time"
 
+	"cloud.google.com/go/compute/metadata"
 	agentendpoint "cloud.google.com/go/osconfig/agentendpoint/apiv1"
-	"github.com/GoogleCloudPlatform/guest-logging-go/logger"
-	"github.com/GoogleCloudPlatform/osconfig/config"
+	"github.com/GoogleCloudPlatform/osconfig/agentconfig"
+	"github.com/GoogleCloudPlatform/osconfig/clog"
 	"github.com/GoogleCloudPlatform/osconfig/retryutil"
 	"github.com/GoogleCloudPlatform/osconfig/tasker"
 	"github.com/GoogleCloudPlatform/osconfig/util"
@@ -34,6 +38,7 @@ import (
 	"google.golang.org/grpc/codes"
 	"google.golang.org/grpc/credentials"
 	"google.golang.org/grpc/status"
+	"google.golang.org/protobuf/proto"
 
 	agentendpointpb "google.golang.org/genproto/googleapis/cloud/osconfig/agentendpoint/v1"
 )
@@ -41,8 +46,11 @@ import (
 const apiRetrySec = 600
 
 var (
-	errServerCancel = errors.New("task canceled by server")
-	taskStateFile   = config.TaskStateFile()
+	errServerCancel      = errors.New("task canceled by server")
+	errServiceNotEnabled = errors.New("service is not enabled for this project")
+	errResourceExhausted = errors.New("ResourceExhausted")
+	taskStateFile        = agentconfig.TaskStateFile()
+	sameStateTimeWindow  = -5 * time.Second
 )
 
 // Client is a an agentendpoint client.
@@ -59,9 +67,9 @@ func NewClient(ctx context.Context) (*Client, error) {
 	opts := []option.ClientOption{
 		option.WithoutAuthentication(), // Do not use oauth.
 		option.WithGRPCDialOption(grpc.WithTransportCredentials(credentials.NewTLS(nil))), // Because we disabled Auth we need to specifically enable TLS.
-		option.WithEndpoint(config.SvcEndpoint()),
+		option.WithEndpoint(agentconfig.SvcEndpoint()),
 	}
-	logger.Debugf("Creating new agentendpoint client.")
+	clog.Debugf(ctx, "Creating new agentendpoint client.")
 	c, err := agentendpoint.NewClient(ctx, opts...)
 	if err != nil {
 		return nil, err
@@ -86,118 +94,160 @@ func (c *Client) Closed() bool {
 	return c.closed
 }
 
+// RegisterAgent calls RegisterAgent discarding the response.
+func (c *Client) RegisterAgent(ctx context.Context) error {
+	token, err := agentconfig.IDToken()
+	if err != nil {
+		return err
+	}
+
+	req := &agentendpointpb.RegisterAgentRequest{AgentVersion: agentconfig.Version(), SupportedCapabilities: agentconfig.Capabilities()}
+	clog.Debugf(ctx, "Calling RegisterAgent with request:\n%s", util.PrettyFmt(req))
+	req.InstanceIdToken = token
+
+	_, err = c.raw.RegisterAgent(ctx, req)
+	return err
+}
+
+// reportInventory calls ReportInventory with the provided inventory.
+func (c *Client) reportInventory(ctx context.Context, inventory *agentendpointpb.Inventory, reportFull bool) (*agentendpointpb.ReportInventoryResponse, error) {
+	token, err := agentconfig.IDToken()
+	if err != nil {
+		return nil, err
+	}
+
+	hash := sha256.New()
+	b, err := proto.Marshal(inventory)
+	if err != nil {
+		return nil, err
+	}
+	io.Copy(hash, bytes.NewReader(b))
+
+	checksum := hex.EncodeToString(hash.Sum(nil))
+	req := &agentendpointpb.ReportInventoryRequest{InventoryChecksum: checksum}
+	if reportFull {
+		req = &agentendpointpb.ReportInventoryRequest{InventoryChecksum: checksum, Inventory: inventory}
+	}
+	clog.Debugf(ctx, "Calling ReportInventory with request:\n%s", util.PrettyFmt(req))
+	req.InstanceIdToken = token
+
+	// Additional logging for verifications in e2e tests.
+	payloadSummary := fmt.Sprintf("hostname %s, short name %s, %d installed packages, %d available packages", inventory.OsInfo.Hostname, inventory.OsInfo.ShortName, len(inventory.InstalledPackages), len(inventory.AvailablePackages))
+	clog.Debugf(ctx, "Calling ReportInventory with request containing %s", payloadSummary)
+
+	return c.raw.ReportInventory(ctx, req)
+}
+
 func (c *Client) startNextTask(ctx context.Context) (res *agentendpointpb.StartNextTaskResponse, err error) {
-	token, err := config.IDToken()
+	token, err := agentconfig.IDToken()
 	if err != nil {
 		return nil, err
 	}
 
 	req := &agentendpointpb.StartNextTaskRequest{}
-	logger.Debugf("Calling StartNextTask with request:\n%s", util.PrettyFmt(req))
+	clog.Debugf(ctx, "Calling StartNextTask with request:\n%s", util.PrettyFmt(req))
 	req.InstanceIdToken = token
 
-	if err := retryutil.RetryAPICall(apiRetrySec*time.Second, "StartNextTask", func() error {
+	if err := retryutil.RetryAPICall(ctx, apiRetrySec*time.Second, "StartNextTask", func() error {
 		res, err = c.raw.StartNextTask(ctx, req)
 		if err != nil {
 			return err
 		}
-		logger.Debugf("StartNextTask response:\n%s", util.PrettyFmt(res))
+		clog.Debugf(ctx, "StartNextTask response:\n%s", util.PrettyFmt(res))
 		return nil
 	}); err != nil {
-		return nil, fmt.Errorf("error calling StartNextTask: %v", err)
+		return nil, fmt.Errorf("error calling StartNextTask: %w", err)
 	}
 
 	return res, nil
 }
 
 func (c *Client) reportTaskProgress(ctx context.Context, req *agentendpointpb.ReportTaskProgressRequest) (res *agentendpointpb.ReportTaskProgressResponse, err error) {
-	token, err := config.IDToken()
+	token, err := agentconfig.IDToken()
 	if err != nil {
 		return nil, err
 	}
 
-	logger.Debugf("Calling ReportTaskProgress with request:\n%s", util.PrettyFmt(req))
+	clog.Debugf(ctx, "Calling ReportTaskProgress with request:\n%s", util.PrettyFmt(req))
 	req.InstanceIdToken = token
 
-	if err := retryutil.RetryAPICall(apiRetrySec*time.Second, "ReportTaskProgress", func() error {
+	if err := retryutil.RetryAPICall(ctx, apiRetrySec*time.Second, "ReportTaskProgress", func() error {
 		res, err = c.raw.ReportTaskProgress(ctx, req)
 		if err != nil {
 			return err
 		}
-		logger.Debugf("ReportTaskProgress response:\n%s", util.PrettyFmt(res))
+		clog.Debugf(ctx, "ReportTaskProgress response:\n%s", util.PrettyFmt(res))
 		return nil
 	}); err != nil {
-		return nil, fmt.Errorf("error calling ReportTaskProgress: %v", err)
+		return nil, fmt.Errorf("error calling ReportTaskProgress: %w", err)
 	}
 
 	return res, nil
 }
 
 func (c *Client) reportTaskComplete(ctx context.Context, req *agentendpointpb.ReportTaskCompleteRequest) error {
-	token, err := config.IDToken()
+	token, err := agentconfig.IDToken()
 	if err != nil {
 		return err
 	}
 
-	logger.Debugf("Calling ReportTaskComplete with request:\n%s", util.PrettyFmt(req))
+	clog.Debugf(ctx, "Calling ReportTaskComplete with request:\n%s", util.PrettyFmt(req))
 	req.InstanceIdToken = token
 
-	if err := retryutil.RetryAPICall(apiRetrySec*time.Second, "ReportTaskComplete", func() error {
+	if err := retryutil.RetryAPICall(ctx, apiRetrySec*time.Second, "ReportTaskComplete", func() error {
 		res, err := c.raw.ReportTaskComplete(ctx, req)
 		if err != nil {
 			return err
 		}
-		logger.Debugf("ReportTaskComplete response:\n%s", util.PrettyFmt(res))
+		clog.Debugf(ctx, "ReportTaskComplete response:\n%s", util.PrettyFmt(res))
 		return nil
 	}); err != nil {
-		return fmt.Errorf("error calling ReportTaskComplete: %v", err)
+		return fmt.Errorf("error calling ReportTaskComplete: %w", err)
 	}
 
 	return nil
 }
 
-// Placeholder for LookupEffectiveGuestPolicies.
-
 func (c *Client) runTask(ctx context.Context) {
-	logger.Debugf("Beginning run task loop.")
+	clog.Debugf(ctx, "Beginning run task loop.")
 	for {
 		res, err := c.startNextTask(ctx)
 		if err != nil {
-			logger.Errorf("Error running StartNextTask, cannot continue: %v", err)
+			clog.Errorf(ctx, "Error running StartNextTask, cannot continue: %v", err)
 			return
 		}
 
 		task := res.GetTask()
 		if task == nil {
-			logger.Debugf("No task to run, ending run task loop.")
+			clog.Debugf(ctx, "No task to run, ending run task loop.")
 			return
 		}
 
-		logger.Debugf("Received task: %s.", task.GetTaskType())
+		clog.Debugf(ctx, "Received task: %s.", task.GetTaskType())
 		switch task.GetTaskType() {
 		case agentendpointpb.TaskType_APPLY_PATCHES:
 			if err := c.RunApplyPatches(ctx, task); err != nil {
-				logger.Errorf("Error running TaskType_APPLY_PATCHES: %v", err)
+				clog.Errorf(ctx, "Error running TaskType_APPLY_PATCHES: %v", err)
 			}
 		case agentendpointpb.TaskType_EXEC_STEP_TASK:
 			if err := c.RunExecStep(ctx, task); err != nil {
-				logger.Errorf("Error running TaskType_EXEC_STEP_TASK: %v", err)
+				clog.Errorf(ctx, "Error running TaskType_EXEC_STEP_TASK: %v", err)
 			}
 		default:
-			logger.Errorf("Unknown task type: %v", task.GetTaskType())
+			clog.Errorf(ctx, "Unknown task type: %v", task.GetTaskType())
 		}
 	}
 }
 
 func (c *Client) handleStream(ctx context.Context, stream agentendpointpb.AgentEndpointService_ReceiveTaskNotificationClient) error {
 	for {
-		logger.Debugf("Waiting on ReceiveTaskNotification stream Recv().")
+		clog.Debugf(ctx, "Waiting on ReceiveTaskNotification stream Recv().")
 		if _, err := stream.Recv(); err != nil {
 			// Return on any stream error, even a close, the caller will simply
 			// reconnect the stream as needed.
 			return err
 		}
-		logger.Debugf("Received task notification.")
+		clog.Debugf(ctx, "Received task notification.")
 
 		// Only queue up one notifcation at a time. We should only ever
 		// have one active task being worked on and one in the queue.
@@ -206,7 +256,7 @@ func (c *Client) handleStream(ctx context.Context, stream agentendpointpb.AgentE
 			// We have been canceled.
 			return nil
 		case c.noti <- struct{}{}:
-			tasker.Enqueue("TaskNotification", func() {
+			tasker.Enqueue(ctx, "TaskNotification", func() {
 				// We lock so that this task will complete before the client can get canceled.
 				c.mx.Lock()
 				defer c.mx.Unlock()
@@ -228,15 +278,15 @@ func (c *Client) handleStream(ctx context.Context, stream agentendpointpb.AgentE
 
 func (c *Client) receiveTaskNotification(ctx context.Context) (agentendpointpb.AgentEndpointService_ReceiveTaskNotificationClient, error) {
 	req := &agentendpointpb.ReceiveTaskNotificationRequest{
-		AgentVersion: config.Version(),
+		AgentVersion: agentconfig.Version(),
 	}
 
-	token, err := config.IDToken()
+	token, err := agentconfig.IDToken()
 	if err != nil {
-		return nil, fmt.Errorf("error fetching Instance IDToken: %v", err)
+		return nil, fmt.Errorf("error fetching Instance IDToken: %w", err)
 	}
 
-	logger.Debugf("Calling ReceiveTaskNotification with request:\n%s", util.PrettyFmt(req))
+	clog.Debugf(ctx, "Calling ReceiveTaskNotification with request:\n%s", util.PrettyFmt(req))
 	req.InstanceIdToken = token
 
 	return c.raw.ReceiveTaskNotification(ctx, req)
@@ -245,11 +295,12 @@ func (c *Client) receiveTaskNotification(ctx context.Context) (agentendpointpb.A
 func (c *Client) loadTaskFromState(ctx context.Context) error {
 	st, err := loadState(taskStateFile)
 	if err != nil {
-		return fmt.Errorf("loadState error: %v", err)
+		return fmt.Errorf("loadState error: %w", err)
 	}
 	if st != nil && st.PatchTask != nil {
 		st.PatchTask.client = c
-		tasker.Enqueue("PatchRun", func() {
+		st.PatchTask.state = st
+		tasker.Enqueue(ctx, "PatchRun", func() {
 			st.PatchTask.run(ctx)
 		})
 	}
@@ -257,9 +308,6 @@ func (c *Client) loadTaskFromState(ctx context.Context) error {
 	return nil
 }
 
-var errServiceNotEnabled = errors.New("service is not enabled for this project")
-var errResourceExhausted = errors.New("ResourceExhausted")
-
 func (c *Client) waitForTask(ctx context.Context) error {
 	stream, err := c.receiveTaskNotification(ctx)
 	if err != nil {
@@ -275,7 +323,7 @@ func (c *Client) waitForTask(ctx context.Context) error {
 		switch s.Code() {
 		case codes.Unavailable:
 			// Something canceled the stream (could be deadline/timeout), we should reconnect.
-			logger.Debugf("Stream canceled, will reconnect: %v", err)
+			clog.Debugf(ctx, "Stream canceled, will reconnect: %v", err)
 			return nil
 		case codes.PermissionDenied:
 			// Service is not enabled for this project.
@@ -284,7 +332,6 @@ func (c *Client) waitForTask(ctx context.Context) error {
 			return errResourceExhausted
 		}
 	}
-	// TODO: Add more error checking (handle more API erros vs non API errors) and backoff where appropriate.
 	return err
 }
 
@@ -297,51 +344,63 @@ func (c *Client) WaitForTaskNotification(ctx context.Context) {
 		// WaitForTaskNotification is already running on this client.
 		return
 	}
-	logger.Debugf("Running WaitForTaskNotification")
+	clog.Debugf(ctx, "Running WaitForTaskNotification")
 	ctx, c.cancel = context.WithCancel(ctx)
 
-	logger.Debugf("Checking local state file for saved task.")
+	clog.Debugf(ctx, "Checking local state file for saved task.")
 	if err := c.loadTaskFromState(ctx); err != nil {
-		logger.Errorf(err.Error())
+		clog.Errorf(ctx, err.Error())
 	}
 
-	logger.Debugf("Setting up ReceiveTaskNotification stream watcher.")
+	clog.Debugf(ctx, "Setting up ReceiveTaskNotification stream watcher.")
 	go func() {
-		resourceExhausted := 1
+		var resourceExhausted int
+		var errs int
+		var sleep time.Duration
 		for {
 			select {
 			case <-ctx.Done():
 				// We have been canceled.
-				logger.Debugf("Disabling WaitForTaskNotification")
+				clog.Debugf(ctx, "Disabling WaitForTaskNotification")
 				return
 			default:
 			}
 
 			if err := c.waitForTask(ctx); err != nil {
-				if err == errServiceNotEnabled {
+				if errors.Is(err, errServiceNotEnabled) {
 					// Service is disabled, close this client and return.
+					clog.Warningf(ctx, "OSConfig Service is disabled.")
 					c.Close()
 					return
 				}
-				logger.Errorf("Error waiting for task: %v", err)
-				sleep := 5 * time.Second
-				if err == errResourceExhausted {
-					sleep = retryutil.RetrySleep(resourceExhausted, 5)
+				var ndr *metadata.NotDefinedError
+				if errors.As(err, &ndr) {
+					// No service account setup for this instance, close this client and return.
+					clog.Warningf(ctx, "No service account set for instance.")
+					c.Close()
+					return
+				}
+
+				if errors.Is(err, errResourceExhausted) {
 					resourceExhausted++
+					sleep = retryutil.RetrySleep(resourceExhausted, 5)
 				} else {
-					resourceExhausted = 1
+					// Retry any other errors with a modest backoff. Only retry up to 10
+					// times, at that point return, the client will be recreated during the next
+					// cycle.
+					errs++
+					clog.Warningf(ctx, "Error waiting for task (attempt %d of 10): %v", errs, err)
+					resourceExhausted = 0
+					if errs > 10 {
+						c.Close()
+						return
+					}
+					sleep = retryutil.RetrySleep(errs, 0)
 				}
 				time.Sleep(sleep)
+				continue
 			}
+			errs = 0
 		}
 	}()
 }
-
-func mkLabels(labels map[string]string) map[string]string {
-	if labels == nil {
-		labels = make(map[string]string)
-	}
-	labels["instance_name"] = config.Name()
-	labels["agent_version"] = config.Version()
-	return labels
-}
