@@ -44,6 +44,7 @@ var (
 	aptGetFullUpgradeCmd = "full-upgrade"
 	aptGetDistUpgradeCmd = "dist-upgrade"
 	aptGetUpgradableArgs = []string{"--just-print", "-qq"}
+	allowDowngradesArg   = "--allow-downgrades"
 
 	dpkgErr = []byte("dpkg --configure -a")
 )
@@ -72,8 +73,9 @@ const (
 )
 
 type aptGetUpgradeOpts struct {
-	upgradeType AptUpgradeType
-	showNew     bool
+	upgradeType     AptUpgradeType
+	showNew         bool
+	allowDowngrades bool
 }
 
 // AptGetUpgradeOption is an option for apt-get upgrade.
@@ -93,6 +95,13 @@ func AptGetUpgradeShowNew(showNew bool) AptGetUpgradeOption {
 	}
 }
 
+// AptGetUpgradeAllowDowngrades returns a AptGetUpgradeOption that specifies AllowDowngrades.
+func AptGetUpgradeAllowDowngrades(allowDowngrades bool) AptGetUpgradeOption {
+	return func(args *aptGetUpgradeOpts) {
+		args.allowDowngrades = allowDowngrades
+	}
+}
+
 func dpkgRepair(ctx context.Context, out []byte) bool {
 	// Error code 100 may occur for non repairable errors, just check the output.
 	if !bytes.Contains(out, dpkgErr) {
@@ -190,8 +199,9 @@ func parseAptUpdates(ctx context.Context, data []byte, showNew bool) []PkgInfo {
 // apt-get [dist-|full-]upgrade.
 func AptUpdates(ctx context.Context, opts ...AptGetUpgradeOption) ([]PkgInfo, error) {
 	aptOpts := &aptGetUpgradeOpts{
-		upgradeType: AptGetUpgrade,
-		showNew:     false,
+		upgradeType:     AptGetUpgrade,
+		showNew:         false,
+		allowDowngrades: false,
 	}
 
 	for _, opt := range opts {
@@ -210,7 +220,7 @@ func AptUpdates(ctx context.Context, opts ...AptGetUpgradeOption) ([]PkgInfo, er
 		return nil, fmt.Errorf("unknown upgrade type: %q", aptOpts.upgradeType)
 	}
 
-	if _, err := run(ctx, aptGet, aptGetUpdateArgs); err != nil {
+	if _, err := AptUpdate(ctx); err != nil {
 		return nil, err
 	}
 
@@ -222,6 +232,11 @@ func AptUpdates(ctx context.Context, opts ...AptGetUpgradeOption) ([]PkgInfo, er
 	return parseAptUpdates(ctx, out, aptOpts.showNew), nil
 }
 
+// AptUpdate runs apt-get update.
+func AptUpdate(ctx context.Context) ([]byte, error) {
+	return run(ctx, aptGet, aptGetUpdateArgs)
+}
+
 func parseInstalledDebpackages(data []byte) []PkgInfo {
 	/*
 	   foo amd64 1.2.3-4
